
% OS, 2007
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% STCtutorial
%%%
%%% This file provides a demonstration of STA/STC analysis of simulated 
%%% neural spike trains.  The examples are based on the following article:
%%%
%%%  ****  JOV ARTICLE *****
%%% Matlab tutorial code based on:
%%% Spike-triggered Neural Characterization. O Schwartz, J W Pillow, N C Rust, 
%%% and E P Simoncelli. Journal of Vision, 2006.
%%%
%%% Do not run this file as a script!  Instead, load it into an editor, 
%%% read  through the comments, and execute groups of consecutive commands
%%% by copy-pasting them into the matlab command window.
%%%
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

startFig=1;      % Starting figure number.  To avoid plotting over 
                 % existing figures, set this to a one greater than the 
                 % current highest figure number.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% I. Generate random stimuli

nFrames = 500000;

rand('state', 0);  % initialize random number generator

xDim = 8;
kernelX = xDim;			% spatial size of filter 
kernelT = 6;			% temporal size of filter 
kernelSize = kernelX * kernelT;

allStim = randn(nFrames, kernelSize);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% II. Gather spiking responses.  Here, we generate these by simulating 
%%% a simple model (choose one from the collection below).  

%% -----------
%% (1) Classic V1 simple cell: single filter followed by power-function nonlinearity 

themodel = 1;

rate = 1/12;
base = 1/200;
p = 2;
th = pi/4;
rate = 1/6;
base = 00;
itau = 1.2;  sig=1.6; per=4.5;
[X,Y] = meshgrid([1:kernelX]-(kernelX+1)/2,[kernelT:-1:1]);

v1 = exp(-X.^2/(2*sig^2)) .* exp(-itau*Y) .* Y.^2 .* ...
     mkSine([kernelT,kernelX],per,pi/4,1,0);

clear modelFilts;
modelFilts(:,:,1) = v1;
v1=vectify(v1)/sqrt(var2(v1));
linResp = base + rate * max((allStim*v1),0).^p;
linResp = 20*linResp/max(linResp);


%% -----------
%% (2) Classic complex cell (Adelson/Bergen '85): sum of squared responses
%% of even- and odd-symmetric filters

themodel = 2;

rate = 1/12;
base = 1/100;
crp = 0.87;

rate = 1/12;
base = 00;
itau = 1.2;  sig=1.6; per=4.5;
[X,Y] = meshgrid([1:kernelX]-(kernelX+1)/2,[kernelT:-1:1]);
v1 = exp(-X.^2/(2*sig^2)) .* exp(-itau*Y) .* Y.^2 .* ...
     mkSine([kernelT,kernelX],per,pi/4,1,0);
v2 = exp(-X.^2/(2*sig^2)) .* exp(-itau*Y) .* Y.^2 .* ...
     mkSine([kernelT,kernelX],per,pi/4,1,pi/2);

clear modelFilts;
modelFilts(:,:,1) = v1;
modelFilts(:,:,2) = v2;
v1=vectify(v1)/sqrt(var2(v1)); v2=vectify(v2)/sqrt(var2(v2));

p=2;
linResp = base + rate*(abs(allStim*v1).^p + abs(allStim*v2).^p);
linResp = linResp/max(linResp);

rot=-pi/3;  % orientation of scatterplot shown
rot = 0;
kX=1;  kY=2;
cont=0.1*[0.5 1 2 4]

%% -----------
%% (3) Divisive normalization model (Heeger '92)

themodel = 3;

rate = 1/12;
base = 1/100;
crp = 0.85;

rate = 1/12;
base = 00;
itau = 1.2;  sig=1.6; per=4.5;
[X,Y] = meshgrid([1:kernelX]-(kernelX+1)/2,[kernelT:-1:1]);
v1 = exp(-X.^2/(2*sig^2)) .* exp(-itau*Y) .* Y.^2 .* ...
     mkSine([kernelT,kernelX],per,pi/4,1,0);
v2 = exp(-X.^2/(2*sig^2)) .* exp(-itau*Y) .* Y.^2 .* ...
     mkSine([kernelT,kernelX],per,pi/4,1,pi/2);
v3 = exp(-X.^2/(2*sig^2)) .* exp(-itau*Y) .* Y.^2 .* ...
    mkSine([kernelT,kernelX],per,pi/2+pi/4,1,0);

clear modelFilts;
modelFilts(:,:,1) = v1;
modelFilts(:,:,2) = v2;
modelFilts(:,:,3) = v3;
v1=vectify(v1)/sqrt(var2(v1)); v2=vectify(v2)/sqrt(var2(v2));
v3=vectify(v3)/sqrt(var2(v3));

rate = .25;
p=2;
l1 = (allStim*v1>0).*((allStim*v1).^2); % half squared
l2 = ((allStim*v2).^2);
l3 = ((allStim*v3).^2);

linResp = (1 + l1) ...
    ./ (1 + .03*l2 + .05*l3);
linResp = 15*rate*linResp/max(linResp);

kX=1;  kY=2;
rot=pi/3;  % orientation of scatterplot shown
rot = 0;
cont=0.1*[0.5 1 2 4];

%% -----------
%% (4) Sum of half squares

themodel = 4;

rate = 1/12;
base = 1/100;
crp = 0.87;

rate = 1/12;
base = 00;
itau = 1.2;  sig=1.6; per=4.5;
[X,Y] = meshgrid([1:kernelX]-(kernelX+1)/2,[kernelT:-1:1]);
v2 = exp(-X.^2/(2*sig^2)) .* exp(-itau*Y) .* Y.^2 .* ...
     mkSine([kernelT,kernelX],per,pi/4,1,pi/2);
v1=shift(reshape(v2,6,8),[0 1]);
clear modelFilts;
modelFilts(:,:,1) = v1;
modelFilts(:,:,2) = v2;
v1=vectify(v1)/sqrt(var2(v1)); v2=vectify(v2)/sqrt(var2(v2));

p=2;
lin1 = allStim*v1; lin1 = (lin1.^p) .* (lin1 > 0);
lin2 = allStim*v2; lin2 = (lin2.^p) .* (lin2 > 0);
linResp = base + rate*(lin1 + lin2);
linResp = linResp/max(linResp);

rot=-pi/3;  % orientation of scatterplot shown
rot = 0;
kX=1;  kY=2;
cont = 0.01*[4,10,18,36,70];
cont=0.1*[0.5 1 2 4]

%% -----------
%%  Display model filter(s) (for any of the above models)

figure(startFig); clf;
subplot(2,2,1);
showIm(reshape(modelFilts(:,:,1),kernelT,kernelX)); 
title('Model filter(s)');
if (size(modelFilts,3) > 1)   % if there's a second filter, show it
  subplot(2,2,2);
  showIm(modelFilts(:,:,2));
end
if (size(modelFilts,3) > 2)   % if there's a third filter, show it
  subplot(2,2,3);
  showIm(modelFilts(:,:,3));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% III.  Estimate linear filters using STA/STC

spikeResp = (linResp > rand(nFrames,1));
spikeResp(1:(kernelT-1)) = 0;  % can't use these
nspikes = sum(spikeResp)
spikeInd=(find(spikeResp>0.5));
spikeCount = zeros(nFrames,1);
spikeCount(spikeInd)=1;
wts = double(spikeResp(spikeInd));
sum(wts)

spikeStim = allStim(spikeInd,:);

%% Compute STA and STC
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[sta,stc] = simpleSTC(allStim, spikeCount, 1);

if 0 % Whiten stimuli

  [sta,stc,rawmu,rawcov] = simpleSTC(allStim, spikeCount, 1);
  % You can try setting the ridge parameter to values above 1,
  % and getting smoother (but biased) results -- easiest to see w/ STA
  ridgeparam = 0;
  covInv = inv(rawcov+ridgeparam*eye(length(rawcov)));
  covInvsqrt = sqrtm(covInv);
  % Whiten STA  -- look at this if using half-rectifying nonlinearity
  sta = covInv*sta;
  % Whiten STC -- if using quadratic nonlinearity
  stc = covInvsqrt*stc*covInvsqrt;  % Compute "whitened" STC

end

nsta = sta./norm(sta);
% Plot STA
figure(startFig+2); clf;
subplot(2,2,1);
showIm(reshape(sta,6,8));
title('sta')

% Project out STA;
stcProj = stc-nsta*nsta'*stc - stc*nsta*nsta' + nsta*nsta'*stc*nsta*nsta';
[v,d] = sortedEig(stcProj);  d = diag(d);

fs = 14;
% Plot the eigenvalues
figure(startFig+3); clf;
subplot(2,2,1);
hold; h2=plot([1:kernelSize], d, 'k.','markerSize',18); hold off
xlabel('eigenvalue number','FontSize',fs);
ylabel('variance','FontSize',fs);

vr = var2(allStim);

%[C, X] = asympEigCum(num, length(spikeInd), vr);
axis([0 48 .2 2.2])

% Plot sample eigenvectors
figure(startFig+4); clf;
subplot(2,2,1);
tmpo=reshape(v(:,1),6,8); subplot(2,2,1); showIm(tmpo); 
title('First eigenvector');
tmpo=reshape(v(:,2),6,8); subplot(2,2,2); showIm(tmpo); 
title('Second eigenvector');
tmpo=reshape(v(:,46),6,8); subplot(2,2,3); showIm(tmpo);
title('Second to last eigenvector');
tmpo=reshape(v(:,47),6,8); subplot(2,2,4); showIm(tmpo);
title('Last eigenvector');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% IV.  Examine nonlinearities

switch themodel
case 1
   basis1 = nsta; basis2 = v(:,45);
case 2
   basis1 = v(:,1); basis2 = v(:,2);
case 3
   basis1 = v(:,47); basis2 = nsta;
case 4
   basis1 = v(:,1); basis2 = nsta;
end

% Projection of all stimuli and spike stimuli onto the axis pair
basis1 = basis1 ./ norm(basis1);
basis2 = basis2 ./ norm(basis2);
basis2_numComp =  basis1 * ...
(basis2' * basis1) / sum(basis1.^2);
basis2 = basis2 - basis2_numComp;
basis2_coords = [norm(basis2_numComp), norm(basis2)]
basis2 = basis2 ./ norm(basis2);
allProj = [allStim*basis2, allStim*basis1];
spikeProj = [spikeStim*basis2, spikeStim*basis1];


mx = 4*sqrt(vr);
binSz = (2*mx)/25;


%% Color and size of scatterplot points
col1='b';  sz1=9;     % raw stimuli
col2='r';  sz2=15;    % spike-triggered stimuli

colbg = 0.6*[1 1 1];		% background color
wht = 0.99*[1 1 1];             % this prints correctly as white!


% 1D projection onto first axis
figure(startFig+5); clf;
project1D(allProj(:,1), spikeProj(:,1), binSz, col1, col2, mx, fs, colbg)

% 1D projection onto second axis
figure(startFig+6); clf;
project1D(allProj(:,2), spikeProj(:,2), binSz, col1, col2, mx, fs, colbg)

% 2D projection onto pair of axes
npts = 6500;
project2D(allProj, spikeProj, binSz, npts, rot, col1, col2, mx, fs, colbg);

